/*
 *      cdsd.c
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <string.h>
#include <sys/poll.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <errno.h>
#include <dirent.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <ctype.h>
#include <grp.h>
#include <sys/time.h>
#include <signal.h>
#include <sys/mount.h>
#include <mntent.h>
#include <gdbm.h>
#include <syslog.h>

#include "cdsd.h"
#include "scsi.h"
#include "udevclient.h"
#include "datalist.h"
#include "serialize.h"
#include "settings.h"
#include "libcdsd.h"

#ifndef INFTIM
   #define INFTIM  -1
#endif


nodeptr devices = NULL;

static volatile sig_atomic_t timer_flag, config_flag, term_flag, pipe_flag;
static volatile uid_t signalsender;

struct cdsd_foreach_data {
	void *ptr;
	unsigned int position;
	unsigned int value;
};

//int cdsd_event_cb(unsigned int major, unsigned int minor);
int cdsd_open_device_rw(struct optical_device *device);
int cdsd_set_speed(struct optical_device *device);
int cdsd_change_speeds(struct optical_device *device, int cdspeed, int dvdspeed, int bdspeed, int polling);
int cdsd_get_device_config(struct optical_device *device);
//int cdsd_scan_devices(struct cdsd_global_settings *settings);
int cdsd_device_unmount(const char *device);
int cdsd_search_devices_list_foreach(void **data, int dindex, void *userdata);
int cdsd_search_udev_list_foreach(void **data, int dindex, void *userdata);
int cdsd_poll_list_foreach(void **data, int dindex, void *userdata);
void cdsd_list_devices_cb(unsigned int *params, unsigned int count, int seq, int size, char **buffer, void *data);
int cdsd_change_speeds_by_id(int id, int online, int *cdspeed, int *dvdspeed, int *bdspeed, int *polling);
char *cdsd_parse_request(struct _cmd *cmd, char *message, int *msglen, int uid, struct cdsd_global_settings *settings);
int cdsd_events_listen(int udevfd, int ctlfd, struct cdsd_global_settings *settings);
int cdsd_config_callback(char *parameter, char *argument, void *userdata);
static void cdsd_signal_handler(int sig, siginfo_t *si, void *ucontext);
static int cdsd_start_daemon(void);
static int cdsd_service_lock(void);
static void cdsd_service_lock_free(int lockfd);


int cdsd_event_cb(unsigned int major, unsigned int minor)
{
	struct cdsd_foreach_data fdata;
	struct optical_device *device;
	
	fdata.position = major;
	fdata.value = minor;
	fdata.ptr = NULL;
	
	datalist_foreach(devices, &fdata, (void *)cdsd_search_udev_list_foreach);
	
	if (fdata.ptr != NULL) {
		device = (struct optical_device *)fdata.ptr;
		cdsd_set_speed(device);
		return 1;
	}
	
	return 0;
}

int cdsd_open_device_rw(struct optical_device *device)
{
	int rwfd = -1;
	struct stat statbuf;
	
	if (device != NULL) {
		if (stat(device->devpath, &statbuf) >= 0) {
			if (S_ISBLK(statbuf.st_mode)) {
				if ((rwfd = open(device->devpath, O_RDWR | O_NONBLOCK)) != -1) {
					return rwfd;
				}
			}
		}
	}
	
	return rwfd;
}

int cdsd_set_speed(struct optical_device *device)
{
	int result = 0;
	int rwfd;
	unsigned short oldspeed, setspeed;
	
	if (device == NULL) return 0;
	
	device->user_state_modified = 0;
	
	if (cdsd_get_device_config(device)) {
		oldspeed = device->cur_read_speed;
		device->media_end_lba = 0;
		if (device->media_exists) {
			//Need to know end LBA to call SET STREAMING command
			if (scsi_get_capacity(device->fd, device) == 0) {
				syslog(LOG_WARNING, "Failed to get end LBA for device %i using default", device->id);
			}
			//Test if disc is blank
			if (scsi_get_disc_information(device->fd, device) == 0) {
				syslog(LOG_ERR, "Failed to get disc information for device %i", device->id);
				return 0;
			}
			//If so, interrupt execution
			if (scsi_test_device_media_blank(device)) {
				syslog(LOG_ERR, "Disc in device %i is blank, nothing to do", device->id);
				return 0;
			}
			
			//Open device in RW mode only if there is media in it
			if ((rwfd = cdsd_open_device_rw(device)) == -1) {
				syslog(LOG_ERR, "Failed to open device %i in RW mode", device->id);
				return 0;
			}
			//SCSI command depends on media type
			if (device->media_type_cd) {
				result = scsi_set_cd_speed(rwfd, device->user_cd_speed, MAXCDSPEED, device);
				if (!result) {
					syslog(LOG_WARNING, "Failed to set CD speed for device %i", device->id);
				}
			} else if (device->media_type_dvd) {
				result = scsi_set_streaming(rwfd, device->user_dvd_speed, MAXDVDSPEED, device);
				if (!result) {
					syslog(LOG_WARNING, "Failed to set DVD speed for device %i", device->id);
				}
			} else if (device->media_type_bd) {
				result = scsi_set_streaming(rwfd, device->user_bd_speed, MAXBDSPEED, device);
				if (!result) {
					syslog(LOG_WARNING, "Failed to set BD speed for device %i", device->id);
				}
			}
			close(rwfd);
		}
	}
	
	if (result) {
		device->user_state_modified = 1;
		if (device->media_type_cd) {
			syslog(LOG_INFO, "Set %ux CD speed for device %i", CDSPEED(device->user_cd_speed), device->id);
		} else if (device->media_type_dvd) {
			syslog(LOG_INFO, "Set %ux DVD speed for device %i", DVDSPEED(device->user_dvd_speed), device->id);
		} else if (device->media_type_bd) {
			syslog(LOG_INFO, "Set %ux BD speed for device %i", BDSPEED(device->user_bd_speed), device->id);
		}
		//Get new speed value from device
		cdsd_get_device_config(device);
		setspeed = device->cur_read_speed;
		//If speed changed, write to syslog
		if (oldspeed != setspeed) {
			if (device->media_type_cd) {
				syslog(LOG_INFO, "Device %i speed changed from %ux to %ux for CD", device->id, CDSPEED(oldspeed), CDSPEED(setspeed));
			} else if (device->media_type_dvd) {
				syslog(LOG_INFO, "Device %i speed changed from %ux to %ux for DVD", device->id, DVDSPEED(oldspeed), DVDSPEED(setspeed));
			} else if (device->media_type_bd) {
				syslog(LOG_INFO, "Device %i speed changed from %ux to %ux for BD", device->id, BDSPEED(oldspeed), BDSPEED(setspeed));
			}
		}
	}
	
	return result;
}

int cdsd_change_speeds(struct optical_device *device, int cdspeed, int dvdspeed, int bdspeed, int polling)
{
	int result = 0;
	
	if (device == NULL) return 0;
	
	if (cdspeed != -1) {
		device->user_cd_speed = CDBYTESTREAM(cdspeed);
		result |= 1;
	}
	
	if (dvdspeed != -1) {
		device->user_dvd_speed = DVDBYTESTREAM(dvdspeed);
		result |= 1;
	}
	
	if (bdspeed != -1) {
		device->user_bd_speed = BDBYTESTREAM(bdspeed);
		result |= 1;
	}
	
	if (polling != -1) {
		if (polling > 0) {
			device->user_polling = 1;
		} else {
			device->user_polling = 0;
		}
		result |= 1;
	}
	
	return result;
}

int cdsd_get_device_config(struct optical_device *device)
{
	int result = 0;
	
	if (device == NULL) return 0;
	
	result = scsi_get_configuration(device->fd, device);
	result = result && scsi_mode_sense(device->fd, device);
	result = result && scsi_get_disc_information(device->fd, device);
	
	return result;
}

int cdsd_scan_devices(struct cdsd_global_settings *settings)
{
	DIR *devdir;
	struct dirent *deviced;
	char devpath[PATH_MAX];
	size_t len;
	struct stat statbuf;
	struct optical_device *device;
	int fd;
	int i = 0;
	GDBM_FILE devdb;
	
	if ((devdir = opendir("/dev")) != NULL) {
		devdb = settings_db_open(SETTINGS_DB_FILENAME);
		while ((deviced = readdir(devdir)) != NULL) {
			if ((deviced->d_name[0] == 's') && ((deviced->d_name[1] == 'r') || (deviced->d_name[1] == 'g')) && (isdigit(deviced->d_name[2]))) {
				len = snprintf((char *)&devpath, PATH_MAX - 1, "/dev/%s", deviced->d_name); 
				devpath[len] = '\0';
				if (stat((char *)&devpath, &statbuf) >= 0) {
					if (S_ISBLK(statbuf.st_mode)) {
						device = calloc(1, sizeof(struct optical_device));
						device->major = major(statbuf.st_rdev);
						device->minor = minor(statbuf.st_rdev);
						
						if ((fd = open((char *)&devpath, O_RDONLY | O_NONBLOCK)) != -1) {
							if (scsi_inquiry(fd, 0, 0, device)) {
								if (datalist_add_node(&devices, (void *)device) != -1) {
									device->id = i;
									device->fd = fd;
									device->devpath = strdup(devpath);
									cdsd_get_device_config(device);
									if (settings == NULL) {
										device->user_cd_speed = MINCDSPEED;
										device->user_dvd_speed = MINDVDSPEED;
										device->user_bd_speed = MINBDSPEED;
										device->user_polling = 1;
										device->user_state_modified = 0;
									} else {
										device->user_cd_speed = settings->default_cd_speed;
										device->user_dvd_speed = settings->default_dvd_speed;
										device->user_bd_speed = settings->default_bd_speed;
										device->user_polling = settings->default_polling;
										device->user_state_modified = 0;
									}
									settings_db_fetch(devdb, device);
									syslog(LOG_INFO, "Device %i added: /dev/%s (%s %s %s) [%i:%i]", device->id, deviced->d_name, device->vendorid, device->productid, device->productver, device->major, device->minor);
									//if we need to set speed right after start
									if (settings->online_speed_change) {
										cdsd_set_speed(device);
										//scsi_read_media_serial_number(device->fd, device);
									}
									i++;
								} else {
									syslog(LOG_WARNING, "Device not added: /dev/%s [%i:%i]", deviced->d_name, device->major, device->minor);
									free(device);
									close(fd);
								}
							}
						} else {
							syslog(LOG_WARNING, "Device not ready: /dev/%s [%i:%i]", deviced->d_name, device->major, device->minor);
							free(device);
							close(fd);
						}
						
					}
				}
				
			
			}
		}
		closedir(devdir);
		settings_db_close(devdb);
	}
	
	return i;
}

int cdsd_device_unmount(const char *device)
{
	FILE *mtab, *ntab;
	struct stat statbuf;
	struct mntent *mnt;
	int mountflag;
	int mountnum = 0;
	int srcfd, endfd;
	int numread;
	char cpbuf[1024];
	
	if (device == NULL) return -1;
	
	if ((stat(device, &statbuf) != -1) && (statbuf.st_mode & S_IFBLK)) {
		if ((mtab = setmntent(MTAB_PATH_ETC, "r")) != NULL) {
			if ((ntab = setmntent(MTAB_PATH_TMP, "w")) != NULL) {
				while ((mnt = getmntent(mtab)) != NULL) {
					mountflag = 1;
					if (strcmp(device, mnt->mnt_fsname) == 0) {
						if ((stat(mnt->mnt_dir, &statbuf) != -1) && (statbuf.st_mode & S_IFDIR)) {
							if (umount(mnt->mnt_dir) == -1) {
								syslog(LOG_ERR, "Error while unmounting device %s from %s", device, mnt->mnt_dir);
							} else {
								mountnum++;
								mountflag = 0;
							}
						}
					}
					
					if (mountflag) {
						addmntent(ntab, mnt);
					}	
				}
				endmntent(ntab);
			}
			endmntent(mtab);
		}
	}
	
	if (mountnum > 0) {
		if ((srcfd = open(MTAB_PATH_TMP, O_RDONLY)) == -1) {
			return mountnum;
		}
		
		if ((endfd = open(MTAB_PATH_ESN, O_CREAT|O_WRONLY|O_TRUNC, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) == -1) {
			close(srcfd);
			return mountnum;
		}
		
		while ((numread = read(srcfd, cpbuf, 1024)) > 0) {
			if (write(endfd, cpbuf, numread) != numread) {
				close(srcfd);
				close(endfd);
				unlink(MTAB_PATH_TMP);
				unlink(MTAB_PATH_ESN);
				syslog(LOG_ERR, "Error while writing mount table");
				return mountnum;
			}
		}
		
		rename(MTAB_PATH_ESN, MTAB_PATH_ETC);
		
		close(srcfd);
		close(endfd);
		unlink(MTAB_PATH_TMP);
	}
	
	return mountnum;
}

int cdsd_search_devices_list_foreach(void **data, int dindex, void *userdata)
{
	struct optical_device *device;
	struct cdsd_foreach_data *fdata;
	
	device = *(struct optical_device **)data;
	fdata = (struct cdsd_foreach_data *)userdata;
	
	if (device->id == fdata->position) {
		fdata->ptr = device;
		return 1;
	}
	
	return 0;
}

int cdsd_search_udev_list_foreach(void **data, int dindex, void *userdata)
{
	struct optical_device *device;
	struct cdsd_foreach_data *fdata;
	
	device = *(struct optical_device **)data;
	fdata = (struct cdsd_foreach_data *)userdata;
	
	if ((device->major == fdata->position) && (device->minor == fdata->value)) {
		fdata->ptr = device;
		return 1;
	}
	
	return 0;
}

int cdsd_poll_list_foreach(void **data, int dindex, void *userdata)
{
	struct optical_device *device;
	int mpoints;
	
	device = *(struct optical_device **)data;
	
	if (device->user_polling) {
		if (scsi_get_event_status_notification(device->fd, 1, device) != -1) {
			if (device->event_eject_request || device->event_media_removal) {
				if ((mpoints = cdsd_device_unmount(device->devpath)) > 0) {
					syslog(LOG_INFO, "Device %i fs unmounted from %i mountpoint(s)", device->id, mpoints);
				}
				cdsd_get_device_config(device);
			}
		} else {
			device->user_polling = 0;
			syslog(LOG_WARNING, "Event polling not supported and disabled on device %u", device->id);
		}
	}
		
	return 0;
}

void cdsd_list_devices_cb(unsigned int *params, unsigned int count, int seq, int size, char **buffer, void *data)
{
	int *ptr;
	struct optical_device *device;
	
	ptr = (int *)data;
	
	device = (struct optical_device *)datalist_get_nth(devices, *ptr);
	
	if (device != NULL) {
		serialize_form_message_buffer(params, count, buffer, MSGOFFSET(seq, size), device->vendorid, device->productid,
										device->devpath, device->id, device->media_profile, device->capabilities_profile,
										CDSPEED(device->user_cd_speed), DVDSPEED(device->user_dvd_speed),
										BDSPEED(device->user_bd_speed), device->user_polling);
	} else {
		serialize_form_message_buffer(params, count, buffer, MSGOFFSET(seq, size), "vendor", "device", "/dev/sr0", 0, 0, 0, 0, 0, 0, 0);
	}
	
	*ptr = *ptr + 1;
}

int cdsd_change_speeds_by_id(int id, int online, int *cdspeed, int *dvdspeed, int *bdspeed, int *polling)
{
	struct cdsd_foreach_data fdata;
	struct optical_device *device;
	GDBM_FILE devdb;
	
	fdata.position = id;
	fdata.ptr = NULL;
	
	datalist_foreach(devices, &fdata, (void *)cdsd_search_devices_list_foreach);
	
	if ((device = (struct optical_device *)fdata.ptr) != NULL) {
		if (cdsd_change_speeds(device, *cdspeed, *dvdspeed, *bdspeed, *polling)) {
			//if we need to set speed right after change
			if (online) cdsd_set_speed(device);
			*cdspeed = CDSPEED(device->user_cd_speed);
			*dvdspeed = DVDSPEED(device->user_dvd_speed);
			*bdspeed = BDSPEED(device->user_bd_speed);
			*polling = device->user_polling;
			//store settings in db
			devdb = settings_db_open(SETTINGS_DB_FILENAME);
			settings_db_store(devdb, device);
			settings_db_close(devdb);
			
			return 1;
		}
	}
	
	return 0;
}

char *cdsd_parse_request(struct _cmd *cmd, char *message, int *msglen, int uid, struct cdsd_global_settings *settings)
{
	int ptr = 0;
	int devnum;
	size_t length;
	char *msg;
	
	char cdsd_list_format[] = CDSD_LIST_FORMAT;
	char cdsd_device_format[] = CDSD_DEVICE_FORMAT;
	char cdsd_request_format[] = CDSD_REQUEST_FORMAT;
	char cdsd_info_format[] = CDSD_INFO_FORMAT;
	char cdsd_capabilities_format[] = CDSD_CAPABILITIES_FORMAT;
	
	unsigned int id, polling;
	signed int cdspeed, dvdspeed, bdspeed;
	struct cdsd_foreach_data fdata;
	struct optical_device *device;	
	
	
	if ((cmd == NULL) || ((cmd != NULL) && (cmd->command == CDSD_COMMAND_UNKNOWN))) {
		//broken command
		cmd->result = CDSD_RESULT_ERROR;
		cmd->error = CDSD_EUNKNOWN;
		*msglen = 0;
		return NULL;
	}
	
	if (settings->validgroup != 0) {
		if (!settings_user_in_group(uid, settings->validgroup)) {
			//user is not in group
			cmd->result = CDSD_RESULT_ERROR;
			cmd->error = CDSD_EACCESS;
			*msglen = 0;
			return NULL;
		}
	}
	
	if (datalist_get_empty(devices)) {
		//no devices found
		cmd->result = CDSD_RESULT_ERROR;
		cmd->error = CDSD_ENODEVICES;
		*msglen = 0;
		return NULL;
	}
	
	if (cmd->command == CDSD_COMMAND_LIST) {
		//list devices
		devnum = datalist_get_length(devices);
		msg = serialize_form_array_message(cdsd_list_format, 0, devnum, &length, cdsd_list_devices_cb, &ptr);
		*msglen = length;
		return msg;
	} else if ((cmd->command == CDSD_COMMAND_SET) && (message != NULL) && (*msglen > 0)) {
		//set speeds for device
		serialize_parse_message((char *)&cdsd_device_format, message, 0, &id, &cdspeed, &dvdspeed, &bdspeed, &polling);
		//find device and change speeds
		if (cdsd_change_speeds_by_id(id, settings->online_speed_change, &cdspeed, &dvdspeed, &bdspeed, &polling)) {
			msg = serialize_form_message((char *)&cdsd_device_format, 0, &length, id, cdspeed, dvdspeed, bdspeed, polling);
			*msglen = length;
			return msg;
		} else {
			//unable to change speed
			cmd->result = CDSD_RESULT_ERROR;
			cmd->error = CDSD_ECOMMANDFAIL;
			*msglen = 0;
			return NULL;
		}
	} else if ((cmd->command == CDSD_COMMAND_INFO) && (message != NULL) && (*msglen > 0)) {
		
		serialize_parse_message((char *)&cdsd_request_format, message, 0, &id);
		
		fdata.position = id;
		fdata.ptr = NULL;
		
		datalist_foreach(devices, &fdata, (void *)cdsd_search_devices_list_foreach);
		
		if ((device = (struct optical_device *)fdata.ptr) != NULL) {
			msg = serialize_form_message((char *)&cdsd_info_format, 0, &length, device->vendorid, device->productid, device->productver, device->productsn,
										device->buffer_size, device->max_read_speed, device->cur_read_speed, device->max_write_speed, device->cur_write_speed,
										device->media_exists, device->media_erasable, device->media_type_cd, device->media_type_dvd, device->media_type_bd);
			printf("INFO: %s, %s, %s, %i, %i, %i\n", device->vendorid, device->productid, device->productver, device->buffer_size, device->max_read_speed, device->cur_read_speed);
			*msglen = length;
			return msg;
		} else {
			//unable to find device
			cmd->result = CDSD_RESULT_ERROR;
			cmd->error = CDSD_ECOMMANDFAIL;
			*msglen = 0;
			return NULL;
		}
	} else if ((cmd->command == CDSD_COMMAND_CAPABILITIES) && (message != NULL) && (*msglen > 0)) {
		
		serialize_parse_message((char *)&cdsd_request_format, message, 0, &id);
		
		fdata.position = id;
		fdata.ptr = NULL;
		
		datalist_foreach(devices, &fdata, (void *)cdsd_search_devices_list_foreach);
		
		if ((device = (struct optical_device *)fdata.ptr) != NULL) {
			msg = serialize_form_message((char *)&cdsd_capabilities_format, 0, &length, device->cap_cd_rom_read, device->cap_cd_r_read, device->cap_cd_rw_read,
										device->cap_dvd_rom_read, device->cap_dvd_ram_read, device->cap_dvd_r_read, device->cap_dvd_r_dl_read, device->cap_dvd_rw_read,
										device->cap_dvd_rw_dl_read, device->cap_dvd_plus_r_read, device->cap_dvd_plus_r_dl_read, device->cap_dvd_plus_rw_read,
										device->cap_dvd_plus_rw_dl_read, device->cap_bd_read, device->cap_bd_r_read, device->cap_bd_re_read, device->cap_hddvd_read,
										device->cap_hddvd_r_read, device->cap_hddvd_r_dl_read, device->cap_hddvd_rw_read, device->cap_hddvd_rw_dl_read);
			*msglen = length;
			return msg;
		} else {
			//unable to find device
			cmd->result = CDSD_RESULT_ERROR;
			cmd->error = CDSD_ECOMMANDFAIL;
			*msglen = 0;
			return NULL;
		}
	} else {
		//unknown command
		cmd->result = CDSD_RESULT_ERROR;
		cmd->error = CDSD_EUNKNOWN;
		*msglen = 0;
		return NULL;
	}
	
	*msglen = 0;
	return NULL;
}

int cdsd_events_listen(int udevfd, int ctlfd, struct cdsd_global_settings *settings)
{
	struct pollfd fds[2];
	char recvbuf[4096];
	int msglen;
	char *newmsg;
	struct ucred credentials;
	struct _cmd cmd;
	int numfds;
	int clifd;
	struct sockaddr_un cliaddr;
	socklen_t cliaddrlen = sizeof(cliaddr);
	sigset_t blockmask, oldmask;
	
	
	fds[0].events = POLLIN;
	fds[0].fd = udevfd;
	fds[1].events = POLLIN;
	fds[1].fd = ctlfd;
	
	sigemptyset(&blockmask);
	sigaddset(&blockmask, SIGALRM);
	sigaddset(&blockmask, SIGHUP);
	sigaddset(&blockmask, SIGPIPE);
		
	while (((numfds = poll((struct pollfd *)&fds, 2, INFTIM)) != -1) || (errno == EINTR)) {
		if (numfds > 0) {
			sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
			if (fds[0].revents & POLLIN) {
				if ((msglen = settings_read_credentials(fds[0].fd, NULL, (char *)&recvbuf, sizeof(recvbuf), &credentials)) > 0) {
					if ((credentials.uid == 0) && (credentials.gid == 0)) {
						udevclient_parse_request((char *)&recvbuf, msglen, cdsd_event_cb);
					}
				}
			} else if (fds[1].revents & POLLIN) {
				if ((clifd = accept(fds[1].fd, &cliaddr, &cliaddrlen)) != -1) {
					if ((msglen = settings_read_credentials(clifd, &cmd, (char *)&recvbuf, sizeof(recvbuf), &credentials)) >= 0) {
						newmsg = cdsd_parse_request(&cmd, (char *)&recvbuf, &msglen, credentials.uid, settings);
						settings_send_credentials(clifd, &cmd, newmsg, msglen);
						if (newmsg != NULL) free(newmsg);
					}
					close(clifd);
				}
			}
			sigprocmask(SIG_SETMASK, &oldmask, NULL);
		} else {
			sigprocmask(SIG_BLOCK, &blockmask, &oldmask);
			if (timer_flag) {
				datalist_foreach(devices, NULL, (void *)cdsd_poll_list_foreach);
				timer_flag = 0;
			} else if (config_flag) {
				if (!settings_user_in_group(signalsender, settings->validgroup)) {
					settings_parse_config_file(SETTINGS_CONFIG_FILE_NAME, NULL, settings, (void *)cdsd_config_callback);
				}
				config_flag = 0;
			} else if (term_flag) {
				term_flag = 0;
				break;
			} else if (pipe_flag) {
				pipe_flag = 0;
				syslog(LOG_WARNING, "Command result not sent to client due to timeout");
			}
			sigprocmask(SIG_SETMASK, &oldmask, NULL);
		}
	}
	
	return 1;
}

static void cdsd_signal_handler(int sig, siginfo_t *si, void *ucontext)
{
	signalsender = si->si_uid;
		
	if (sig == SIGALRM) {
		timer_flag = 1;
	} else if (sig == SIGHUP) {
		config_flag = 1;
	} else if ((sig == SIGTERM) || (sig == SIGINT)) {
		term_flag = 1;
	} else if (sig == SIGPIPE) {
		pipe_flag = 1;
	}
}

int cdsd_config_callback(char *parameter, char *argument, void *userdata)
{
	gid_t group;
	unsigned short value;
	struct cdsd_global_settings *settings;
	
	settings = (struct cdsd_global_settings *)userdata;
	
	if (strcmp(parameter, "groupname") == 0) {
		strncpy((char *)&settings->validgroupname, argument, sizeof(settings->validgroupname));
		if ((group = settings_get_group_id(settings->validgroupname)) != -1) settings->validgroup = group;
	} else if (strcmp(parameter, "cdspeed") == 0) {
		value = CDBYTESTREAM((unsigned short)strtoul(argument, (char **)NULL, 10));
		if (value > MAXCDSPEED) settings->default_cd_speed = MAXCDSPEED; else settings->default_cd_speed = value;
	} else if (strcmp(parameter, "dvdspeed") == 0) {
		value = DVDBYTESTREAM((unsigned short)strtoul(argument, (char **)NULL, 10));
		if (value > MAXDVDSPEED) settings->default_dvd_speed = MAXDVDSPEED; else settings->default_dvd_speed = value;
	} else if (strcmp(parameter, "bdspeed") == 0) {
		value = BDBYTESTREAM((unsigned short)strtoul(argument, (char **)NULL, 10));
		if (value > MAXBDSPEED) settings->default_bd_speed = MAXBDSPEED; else settings->default_bd_speed = value;
	} else if (strcmp(parameter, "polling") == 0) {
		value = (unsigned short)strtoul(argument, (char **)NULL, 10);
		if (value) settings->default_polling = 1; else settings->default_polling = 0;
	} else if (strcmp(parameter, "online_speed_change") == 0) {
		value = (unsigned short)strtoul(argument, (char **)NULL, 10);
		if (value) settings->online_speed_change = 1; else settings->online_speed_change = 0;
	}
		
	return 0;
}

static int cdsd_start_daemon(void)
{
	int maxfd, fd;
	struct sigaction sa;
	
	
	switch (fork()) {
		case -1: return -1;
		case  0: break;
		default: _exit(EXIT_SUCCESS);
	}
	
	setsid();
	
	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	
	sigaction(SIGHUP, &sa, NULL);
		
	switch (fork()) {
		case -1: return -1;
		case  0: break;
		default: _exit(EXIT_SUCCESS);
	}
	
	umask(0);
	
	chdir("/");
	
	if ((maxfd = sysconf(_SC_OPEN_MAX)) == -1) maxfd = 1024;
	
	for (fd = 0; fd < maxfd; fd++) close(fd);
	
	close(STDIN_FILENO);
	
	if ((fd = open("/dev/null", O_RDWR)) != STDIN_FILENO) return -1;
	if (dup2(STDIN_FILENO, STDOUT_FILENO) != STDOUT_FILENO) return -1;
	if (dup2(STDIN_FILENO, STDERR_FILENO) != STDERR_FILENO) return -1;
	
	return 0;
}

static int cdsd_service_lock(void)
{
	int fd;
	char buffer[16];
	struct flock lock;
	
	if ((fd = open(SETTINGS_LOCK_FILE_NAME, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) < 0) {
		return -1;
	}
	
	lock.l_type = F_WRLCK;
	lock.l_whence = SEEK_SET;
	lock.l_start = 0;
	lock.l_len = 0;
	
	if (fcntl(fd, F_SETLK, &lock) == -1) {
		if (errno == EACCES || errno == EAGAIN) {
			close(fd);
			return -1;
		}
	}
	
	ftruncate(fd, 0);
	memset(buffer, 0, sizeof(buffer));
	sprintf(buffer, "%ld", (long)getpid());
	write(fd, buffer, strlen(buffer));
	
	return fd;
}

static void cdsd_service_lock_free(int lockfd)
{
	close(lockfd);
	unlink(SETTINGS_LOCK_FILE_NAME);
}

int main(int argc, char **argv)
{
	int opt, daemonize;
	int udevfd, ctlfd;
	int lockfd;
	struct itimerval timer;
	struct sigaction sa;
	struct cdsd_global_settings settings;
	
	
	daemonize = 0;
	
	while ((opt = getopt(argc, argv, "dh")) != -1) {
		switch (opt) {
			case 'd':
				daemonize = 1;
				break;
			case 'h':
				printf("cdsd [-d] [-h]\n\n"
						"COMMANDS:\n"
						"  -d		Run as daemon\n"
						"  -h		Print help and exit\n");
				exit(EXIT_SUCCESS);
			default:
				fprintf(stderr, "Usage: %s [-d] [-h]\n", argv[0]);
				exit(EXIT_FAILURE);
		}
	}
	
	if (daemonize) cdsd_start_daemon();
		
	settings.default_cd_speed = MINCDSPEED;
	settings.default_dvd_speed = MINDVDSPEED;
	settings.default_bd_speed = MINBDSPEED;
	settings.default_polling = 1;
	settings.online_speed_change = 0;
	
	openlog("cdsd", LOG_CONS|LOG_PERROR, LOG_USER);
	
	if ((lockfd = cdsd_service_lock()) == -1) {
		syslog(LOG_ERR, "Service already running");
		exit(EXIT_FAILURE);
	}
	
	strncpy((char *)&settings.validgroupname, "cdrom", sizeof(settings.validgroupname));
	if ((settings.validgroup = settings_get_group_id("cdrom")) == -1) {
		syslog(LOG_WARNING, "Group \"cdrom\" not found, everybody can control service");
		settings.validgroup = 0;
	}
	
	settings_parse_config_file(SETTINGS_CONFIG_FILE_NAME, NULL, &settings, (void *)cdsd_config_callback);
	
	if (!cdsd_scan_devices(&settings)) {
		syslog(LOG_EMERG, "No devices found, exiting");
		closelog();
		exit(EXIT_FAILURE);
	}
	
	if ((udevfd = udevclient_connection_create()) == -1) {
		syslog(LOG_EMERG, "Failed to subscribe to udev events");
		closelog();
		exit(EXIT_FAILURE);
	}
	
	if ((ctlfd = settings_server_connection_create()) == -1) {
		syslog(LOG_EMERG, "Failed to bind control socket");
		closelog();
		close(udevfd);
		exit(EXIT_FAILURE);
	}
	
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_INTERRUPT | SA_SIGINFO;
	sa.sa_sigaction = cdsd_signal_handler;
	
	if (sigaction(SIGALRM, &sa, NULL) == -1) {
		syslog(LOG_ERR, "Failed to start event polling");
	}
	
	if (sigaction(SIGHUP, &sa, NULL) == -1) {
		syslog(LOG_ERR, "Failed to enable SIGHUP handler");
	}
	
	if (sigaction(SIGTERM, &sa, NULL) == -1) {
		syslog(LOG_ERR, "Failed to enable SIGTERM handler");
	}
	
	if (sigaction(SIGINT, &sa, NULL) == -1) {
		syslog(LOG_ERR, "Failed to enable SIGINT handler");
	}
	
	if (sigaction(SIGPIPE, &sa, NULL) == -1) {
		syslog(LOG_ERR, "Failed to enable SIGPIPE handler");
	}
	
	timer_flag = 0;
	config_flag = 0;
	term_flag = 0;
	pipe_flag = 0;
	
	timer.it_value.tv_sec =  3;
	timer.it_value.tv_usec = 0;
	timer.it_interval.tv_sec =  3;
	timer.it_interval.tv_usec = 0;
	
	setitimer(ITIMER_REAL, &timer, NULL);
	
	cdsd_events_listen(udevfd, ctlfd, &settings);
	
	closelog();
	
	settings_server_connection_close(ctlfd);
	
	cdsd_service_lock_free(lockfd);
	
	return EXIT_SUCCESS;
}
