/*
 *      cdsdctl.c
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <dlfcn.h>

#include "libcdsd.h"
#include "resources.h"

enum {
	INFO_DATA_STRING = 0,
	INFO_DATA_BYTE,
	INFO_DATA_SHORT,
	INFO_DATA_INT
};

typedef struct {
	const char *description;
	const char *format;
	const int datatype;
	void *value;
} cdsd_info_table;

//CDSD library functions
int  (*libcdsd_command_list_dl)(int *error, libcdsd_list_cb list_cb);
int  (*libcdsd_command_modeset_dl)(int *error, struct cdsd_settings_device *device);
int  (*libcdsd_command_info_dl)(int *error, int deviceid, struct cdsd_settings_info *info);
int  (*libcdsd_command_capabilities_dl)(int *error, int deviceid, struct cdsd_capabilities_info *capabilities);
int  (*libcdsd_connection_available_dl)(int *error);
/*int  (*libcdsd_get_error_dl)(void);*/

int cdsdctl_load_library(void)
{
	void *cdsdlib;
	char *liberror;
	int i;
	
	typedef struct {
		void **function;
		const char *alias;
	} library;
	
	library cdsdcore[] = {
		{(void **)(&libcdsd_command_list_dl), "libcdsd_command_list"},
		{(void **)(&libcdsd_command_modeset_dl), "libcdsd_command_modeset"},
		{(void **)(&libcdsd_command_info_dl), "libcdsd_command_info"},
		{(void **)(&libcdsd_command_capabilities_dl), "libcdsd_command_capabilities"},
		{(void **)(&libcdsd_connection_available_dl), "libcdsd_connection_available"},
		/*{(void **)(&libcdsd_get_error_dl), "libcdsd_get_error"},*/
		{NULL, NULL}
	};
	
	
	if ((cdsdlib = dlopen(RESOURCE_CDSD_LIBRARY, RTLD_LAZY)) == NULL) {
		fprintf(stderr, "Failed to open shared library: %s\n", dlerror());
		return -1;
	}
	
	dlerror();
	
	i = 0;
	while (cdsdcore[i].alias != NULL) {
		*(cdsdcore[i].function) = dlsym(cdsdlib, cdsdcore[i].alias); 
		if ((liberror = dlerror()) != NULL) {
			fprintf(stderr, "Failed to assign library function \"%s\": %s\n", cdsdcore[i].alias, liberror);
			return -1;
		}
		i++;
	}
	
	return 0;
}

void cdsdctl_print_error(int error, char *str)
{
	char *errstr;
	
	switch (error) {
		case CDSD_EINTERNAL:
			errstr = "Internal error";
			break;
		case CDSD_EACCESS:
			errstr = "No access permission\nAre you memeber of 'cdrom' group?\nTry running 'usermod -a -G cdrom <username>' as root.";
			break;
		case CDSD_ENODEVICES:
			errstr = "No devices available";
			break;
		case CDSD_ECOMMANDFAIL:
			errstr = "Command execution fail";
			break;
		case CDSD_ENOSERVICE:
			errstr = "Service not running";
			break;
		case CDSD_EREQUESTFAIL:
			errstr = "Wrong request";
			break;
		default:
			errstr = "Unknown error";
			break;
	}
	
	if (str == NULL) {
		fprintf(stderr, "%s\n", errstr);
	} else {
		fprintf(stderr, "%s: %s\n", str, errstr);
	}
}

int cdsdctl_devices_list_cb(struct cdsd_settings_device *device)
{
	printf("%4u%8s %16s%16s%5ux%5ux%5ux%5u\n", device->id, device->vendor, device->product, device->fname,
										device->cdspeed, device->dvdspeed, device->bdspeed, device->polling);
	
	return 0;
}

int cdsdctl_list_devices(void)
{
	int numdevs, error;
	
	printf("Available devices:\n");
	printf("\033[7m%4s%25s%16s%6s%6s%6s%5s\033[0m\n", "ID", "Device", "File", "CD", "DVD", "BD", "POLL");
	if ((numdevs = libcdsd_command_list_dl(&error, cdsdctl_devices_list_cb)) < 1) {
		cdsdctl_print_error(error, "No devices available");
	}
	
	return numdevs;
}

int cdsdctl_set_speed(unsigned int devid, int cd, int dvd, int bd, int poll)
{
	int numdevs, error;
	struct cdsd_settings_device device;
	
	device.id = devid;
	device.cdspeed = cd;
	device.dvdspeed = dvd;
	device.bdspeed = bd;
	device.polling = poll;
	
	if ((numdevs = libcdsd_command_modeset_dl(&error, &device)) < 1) {
		cdsdctl_print_error(error, "Failed to set speeds for device");
	} else {
		printf("New device speeds:\n");
		printf("\033[7m%4s%6s%6s%6s%5s\033[0m\n", "ID", "CD", "DVD", "BD", "POLL");
		printf("%4u%5ux%5ux%5ux%5u\n", device.id, device.cdspeed, device.dvdspeed, device.bdspeed, device.polling);
	}
	
	return numdevs;
}

int cdsdctl_device_info(int devid)
{
	struct cdsd_settings_info info;
	int error, i;
	char buffer[32];
	
	cdsd_info_table infotable[] = {
		{"Vendor",              "%s",      INFO_DATA_STRING, (void *)(&info.vendorid)},
		{"Product",             "%s",      INFO_DATA_STRING, (void *)(&info.productid)},
		{"Version",             "%s",      INFO_DATA_STRING, (void *)(&info.productver)},
		{"Serial",              "%s",      INFO_DATA_STRING, (void *)(&info.productsn)},
		{"Buffer size",         "%u kb",   INFO_DATA_SHORT,  (void *)(&info.buffer_size)},
		{"Maximum read speed",  "%u kb/s", INFO_DATA_SHORT,  (void *)(&info.max_read_speed)},
		{"Current read speed",  "%u kb/s", INFO_DATA_SHORT,  (void *)(&info.cur_read_speed)},
		{"Maximum write speed", "%u kb/s", INFO_DATA_SHORT,  (void *)(&info.max_write_speed)},
		{"Current write speed", "%u kb/s", INFO_DATA_SHORT,  (void *)(&info.cur_write_speed)},
		{"Media inserted",      "%u",      INFO_DATA_BYTE,   (void *)(&info.media_exists)},
		{"Media is eraseable",  "%u",      INFO_DATA_BYTE,   (void *)(&info.media_erasable)},
		{"Media is CD",         "%u",      INFO_DATA_BYTE,   (void *)(&info.media_type_cd)},
		{"Media is DVD",        "%u",      INFO_DATA_BYTE,   (void *)(&info.media_type_dvd)},
		{"Media is BD",         "%u",      INFO_DATA_BYTE,   (void *)(&info.media_type_bd)},
		{NULL,                  NULL,      INFO_DATA_INT,    NULL}
	};
	
		
	if (libcdsd_command_info_dl(&error, devid, &info) < 1) {
		cdsdctl_print_error(error, "Failed to get info for device");
		return 0;
	} else {
		i = 0;
		printf("\033[7m%25s%25s\033[0m\n", "Parameter", "Value");
		while (infotable[i].value != NULL) {
			if (infotable[i].datatype == INFO_DATA_STRING) {
				snprintf((char *)&buffer, sizeof(buffer), infotable[i].format, (char *)(infotable[i].value));
			} else if (infotable[i].datatype == INFO_DATA_SHORT) {
				snprintf((char *)&buffer, sizeof(buffer), infotable[i].format, *(unsigned short *)(infotable[i].value));
			} else if (infotable[i].datatype == INFO_DATA_BYTE) {
				snprintf((char *)&buffer, sizeof(buffer), infotable[i].format, *(unsigned char *)(infotable[i].value));
			}
			
			printf("%25s%25s\n", infotable[i].description, buffer);
			
			i++;
		}
	}
	
	return 1;
}

int cdsdctl_device_capabilities(int devid)
{
	struct cdsd_capabilities_info capabilities;
	int error, i;
	char buffer[32];
	
	cdsd_info_table infotable[] = {
		{"AUDIO CD",        "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_cdrom)},
		{"CD-R",            "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_cd_r)},
		{"CD-RW",           "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_cd_rw)},
		{"DVD",             "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_rom)},
		{"DVDRAM",          "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_ram)},
		{"DVD-R",           "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_r)},
		{"DVD-R DL",        "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_r_dl)},
		{"DVD-RW",          "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_rw)},
		{"DVD-RW DL",       "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_rw_dl)},
		{"DVD+R",           "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_plus_r)},
		{"DVD+R DL",        "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_plus_r_dl)},
		{"DVD+RW",          "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_plus_rw)},
		{"DVD+RW DL",       "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_dvd_plus_rw_dl)},
		{"BD",              "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_bd_rom)},
		{"BD-R",            "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_bd_r)},
		{"BD-RE",           "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_bd_re)},
		{"HDDVD",           "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_hddvd_rom)},
		{"HDDVD-R",         "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_hddvd_r)},
		{"HDDVD-R DL",      "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_hddvd_r_dl)},
		{"HDDVD-RW",        "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_hddvd_rw)},
		{"HDDVD-RW DL",     "%s",      INFO_DATA_STRING, (void *)(&capabilities.cap_hddvd_rw_dl)},
		{NULL,              NULL,      INFO_DATA_INT,    NULL}
	};
	
		
	if (libcdsd_command_capabilities_dl(&error, devid, &capabilities) < 1) {
		cdsdctl_print_error(error, "Failed to get capabilities info for device");
		return 0;
	} else {
		i = 0;
		printf("\033[7m%25s%25s\033[0m\n", "Parameter", "Value");
		while (infotable[i].value != NULL) {
			if (infotable[i].datatype == INFO_DATA_STRING) {
				if (*(unsigned char *)(infotable[i].value)) {
					snprintf((char *)&buffer, sizeof(buffer), "%s", "YES");
				} else {
					snprintf((char *)&buffer, sizeof(buffer), "%s", "NO");
				}
				printf("%25s%25s\n", infotable[i].description, buffer);
			}
			i++;
		}
	}
	
	return 1;
}

void cdsdctl_print_help(void)
{
	fprintf(stderr, "Usage: cdsdctl [-l] [-g] [-a] [-i device] [-c cdspeed] [-d dvdspeed] [-b bdspeed] [-p polling]\n");
	fprintf(stderr, " -l - list devices\n -g - get device info\n -a - get device capabilities\n");
	exit(EXIT_FAILURE);
}

int main(int argc, char **argv)
{
	int lsflag, speedflag, infoflag, capsflag, opt, error;
	unsigned long int devid;
	long int cdspeed, dvdspeed, bdspeed, polling;
	
	lsflag = 0;
	speedflag = 0;
	infoflag = 0;
	capsflag = 0;
	devid = 0;
	cdspeed = -1;
	dvdspeed = -1;
	bdspeed = -1;
	polling = -1;
		
	while ((opt = getopt(argc, argv, "lgai:c:d:b:p:")) != -1) {
		switch (opt) {
			case 'l':
				lsflag = 1;
				break;
			case 'g':
				infoflag = 1;
				break;
			case 'a':
				capsflag = 1;
				break;
			case 'i':
				devid = strtoul(optarg, (char **)NULL, 10);
				break;
			case 'c':
				cdspeed = strtol(optarg, (char **)NULL, 10);
				if (cdspeed < -1) cdspeed = -1;
				if (cdspeed > 56) cdspeed = 56;
				speedflag |= 1;
				break;
			case 'd':
				dvdspeed = strtol(optarg, (char **)NULL, 10);
				if (dvdspeed < -1) dvdspeed = -1;
				if (dvdspeed > 24) dvdspeed = 24;
				speedflag |= 1;
				break;
			case 'b':
				bdspeed = strtol(optarg, (char **)NULL, 10);
				if (bdspeed < -1) bdspeed = -1;
				if (bdspeed > 12) bdspeed = 12;
				speedflag |= 1;
				break;
			case 'p':
				polling = strtol(optarg, (char **)NULL, 10);
				if (polling < -1) polling = -1;
				if (polling > 1) polling = 1;
				speedflag |= 1;
				break;
			default: /* '?' */
				cdsdctl_print_help();
		}
	}
	
	if (argc == 1) lsflag = 1;
	
	if (cdsdctl_load_library() == -1) {
		fprintf(stderr, "Unable to load shared library\n");
		exit(EXIT_FAILURE);
	}
	
	if (libcdsd_connection_available_dl(&error) == -1) {
		fprintf(stderr, "CDSD service not running\n");
		exit(EXIT_FAILURE);
	}
	
	if (lsflag) {
		cdsdctl_list_devices();
	} else if (infoflag) {
		cdsdctl_device_info(devid);
	} else if (capsflag) {
		cdsdctl_device_capabilities(devid);
	} else if (speedflag) {
		cdsdctl_set_speed(devid, cdspeed, dvdspeed, bdspeed, polling);
	} else {
		cdsdctl_print_help();
    }
	
	return 0;
}
