/*
 *      libcdsd.c
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>

#include "settings.h"
#include "libcdsd.h"
#include "serialize.h"
#include "datalist.h"

//static cdsderror = CDSD_ESUCCESS;


void _init(void)
{
	//printf("init\n");
}

void _fini(void)
{
	//printf("fini\n");
}

int libcdsd_send_request(int cdsdfd, struct _cmd *cmd, char *data, size_t size, size_t fullsize)
{
	struct pollfd fds[1];
	struct ucred credentials;
	
	fds[0].events = POLLIN;
	fds[0].fd = cdsdfd;
	
	if (cmd == NULL) return -1;
	
	if (settings_send_credentials(cdsdfd, cmd, data, size) == -1) return -1;
	
	if (poll((struct pollfd *)&fds, 1, CDSD_ANSWER_TIMEOUT) != -1) {
		if (fds[0].revents & POLLIN) {
			if ((size = settings_read_credentials(fds[0].fd, cmd, data, fullsize, &credentials)) > 0) {
				return size;			
			}
		}
	}
	
	return -1;
}

void libcdsd_command_list_cb(unsigned int *params, unsigned int count, int seq, int size, char **message, void *data)
{
	nodeptr *devices;
	struct cdsd_settings_device *device;
	
	devices = (nodeptr *)data;
	
	if ((device = calloc(1, sizeof(struct cdsd_settings_device))) == NULL) {
		//fprintf(stderr, "Failed to allocate memory\n");
		return;
	}
	
	serialize_parse_message_buffer(params, count, message, MSGOFFSET(seq, size), &(device->vendor), &(device->product), &(device->fname),
									&device->id, &device->media, &device->capabilities,	&device->cdspeed, &device->dvdspeed, &device->bdspeed,
									&device->polling);
	
	if (datalist_add_node(devices, (void *)device) == -1) {
		free(device);
		//fprintf(stderr, "Failed to add device into linked list\n");
		return;
	}
}

int cdsd_poll_list_foreach(void **data, int dindex, void *userdata)
{
	libcdsd_list_cb list_cb;
	struct cdsd_settings_device *device;
	
	list_cb = *(libcdsd_list_cb *)userdata;
	device = *(struct cdsd_settings_device **)data;
	
	list_cb(device);
	
	return 0;
}

int libcdsd_command_list(int *error, libcdsd_list_cb list_cb)
{
	int sockfd;
	struct _cmd cmd;
	nodeptr devices = NULL;
	char buffer[4096*4];
	size_t datasize;
	int datalen;
	char cdsd_list_format[] = CDSD_LIST_FORMAT;
	char acmd;
	unsigned int numdevs;
		
	if (list_cb == NULL) return -1;
	
	if ((sockfd = settings_client_connection_create()) == -1) {
		if (error != NULL) *error = CDSD_ENOSERVICE;
		return -1;
	}
	
	cmd.command = CDSD_COMMAND_LIST;
	cmd.result = CDSD_RESULT_SUCCESS;
	cmd.error = CDSD_EUNKNOWN;
	datasize = 0;
	
	if ((datalen = libcdsd_send_request(sockfd, &cmd, (char *)&buffer, datasize, sizeof(buffer))) <= 0) {
		if (error != NULL) *error = cmd.error;
		close(sockfd);
		return -1;
	}
	
	if ((cmd.result == CDSD_RESULT_SUCCESS) && (datalen > 0)) {
		serialize_parse_array_message(cdsd_list_format, &numdevs, (char *)&buffer, &acmd, libcdsd_command_list_cb, &devices);
		datalist_foreach(devices, &list_cb, (void *)cdsd_poll_list_foreach);
	}
	
	if (error != NULL) *error = CDSD_ESUCCESS;
	close(sockfd);
	
	return numdevs;
}

int libcdsd_command_modeset(int *error, struct cdsd_settings_device *device)
{
	int sockfd;
	struct _cmd cmd;
	char buffer[4096*4];
	size_t datasize;
	int numdevs;
	char cdsd_device_format[] = CDSD_DEVICE_FORMAT;
	char *msg;
	
	
	if (device == NULL) return -1;
	
	if ((sockfd = settings_client_connection_create()) == -1) {
		if (error != NULL) *error = CDSD_ENOSERVICE;
		return -1;
	}
		
	cmd.command = CDSD_COMMAND_SET;
	cmd.result = CDSD_RESULT_SUCCESS;
	cmd.error = CDSD_EUNKNOWN;
	
	if ((msg = serialize_form_message((char *)&cdsd_device_format, 0, &datasize, device->id, device->cdspeed, device->dvdspeed, device->bdspeed, device->polling)) == NULL) {
		if (error != NULL) *error = CDSD_EINTERNAL;
		return -1;
	}
	
	memcpy((char *)&buffer, msg, datasize);
	free(msg);
	
	if ((datasize = libcdsd_send_request(sockfd, &cmd, (char *)&buffer, datasize, sizeof(buffer))) == -1) {
		if (error != NULL) *error = cmd.error;
		close(sockfd);
		return -1;
	}
	
	if ((cmd.result == CDSD_RESULT_SUCCESS) && (datasize > 0)) {
		serialize_parse_message((char *)&cdsd_device_format, (char *)&buffer, 0, &device->id, &device->cdspeed, &device->dvdspeed, &device->bdspeed, &device->polling);
		numdevs = 1;
	}
	
	if (error != NULL) *error = CDSD_ESUCCESS;
	close(sockfd);
	
	return numdevs;
}

int libcdsd_command_info(int *error, int deviceid, struct cdsd_settings_info *info)
{
	int sockfd;
	struct _cmd cmd;
	char buffer[4096*4];
	size_t datasize;
	int res;
	char cdsd_request_format[] = CDSD_REQUEST_FORMAT;
	char cdsd_info_format[] = CDSD_INFO_FORMAT;
	char *msg;
	
	if (info == NULL) return -1;
	
	if ((sockfd = settings_client_connection_create()) == -1) {
		if (error != NULL) *error = CDSD_ENOSERVICE;
		return -1;
	}
	
	cmd.command = CDSD_COMMAND_INFO;
	cmd.result = CDSD_RESULT_SUCCESS;
	cmd.error = CDSD_EUNKNOWN;
	
	if ((msg = serialize_form_message((char *)&cdsd_request_format, 0, &datasize, deviceid)) == NULL) {
		if (error != NULL) *error = CDSD_EINTERNAL;
		return -1;
	}
	
	memcpy((char *)&buffer, msg, datasize);
	free(msg);
	
	if ((datasize = libcdsd_send_request(sockfd, &cmd, (char *)&buffer, datasize, sizeof(buffer))) == -1) {
		if (error != NULL) *error = cmd.error;
		close(sockfd);
		return -1;
	}
	
	memset(info, 0, sizeof(struct cdsd_settings_info));
	
	if ((cmd.result == CDSD_RESULT_SUCCESS) && (datasize > 0)) {
		serialize_parse_message_static((char *)&cdsd_info_format, (char *)&buffer, 0, &info->vendorid, &info->productid, &info->productver, &info->productsn,
										&info->buffer_size, &info->max_read_speed, &info->cur_read_speed, &info->max_write_speed, &info->cur_write_speed,
										&info->media_exists, &info->media_erasable, &info->media_type_cd, &info->media_type_dvd, &info->media_type_bd);
		res = 1;
	} else {
		res = 0;
	}
	
	if (error != NULL) *error = CDSD_ESUCCESS;
	close(sockfd);
	
	return res;
}

int libcdsd_command_capabilities(int *error, int deviceid, struct cdsd_capabilities_info *capabilities)
{
	int sockfd;
	struct _cmd cmd;
	char buffer[4096*4];
	size_t datasize;
	int res;
	char cdsd_request_format[] = CDSD_REQUEST_FORMAT;
	char cdsd_capabilities_format[] = CDSD_CAPABILITIES_FORMAT;
	char *msg;
	
	if (capabilities == NULL) return -1;
	
	if ((sockfd = settings_client_connection_create()) == -1) {
		if (error != NULL) *error = CDSD_ENOSERVICE;
		return -1;
	}
	
	cmd.command = CDSD_COMMAND_CAPABILITIES;
	cmd.result = CDSD_RESULT_SUCCESS;
	cmd.error = CDSD_EUNKNOWN;
	
	if ((msg = serialize_form_message((char *)&cdsd_request_format, 0, &datasize, deviceid)) == NULL) {
		if (error != NULL) *error = CDSD_EINTERNAL;
		return -1;
	}
	
	memcpy((char *)&buffer, msg, datasize);
	free(msg);
	
	if ((datasize = libcdsd_send_request(sockfd, &cmd, (char *)&buffer, datasize, sizeof(buffer))) == -1) {
		if (error != NULL) *error = cmd.error;
		close(sockfd);
		return -1;
	}
	
	memset(capabilities, 0, sizeof(struct cdsd_capabilities_info));
	
	if ((cmd.result == CDSD_RESULT_SUCCESS) && (datasize > 0)) {
		serialize_parse_message_static((char *)&cdsd_capabilities_format, (char *)&buffer, 0, &capabilities->cap_cdrom, &capabilities->cap_cd_r, &capabilities->cap_cd_rw,
										&capabilities->cap_dvd_rom,	&capabilities->cap_dvd_ram, &capabilities->cap_dvd_r, &capabilities->cap_dvd_r_dl,
										&capabilities->cap_dvd_rw, &capabilities->cap_dvd_rw_dl, &capabilities->cap_dvd_plus_r, &capabilities->cap_dvd_plus_r_dl,
										&capabilities->cap_dvd_plus_rw, &capabilities->cap_dvd_plus_rw_dl, &capabilities->cap_bd_rom, &capabilities->cap_bd_r,
										&capabilities->cap_bd_re, &capabilities->cap_hddvd_rom, &capabilities->cap_hddvd_r, &capabilities->cap_hddvd_r_dl,
										&capabilities->cap_hddvd_rw, &capabilities->cap_hddvd_rw_dl);
		res = 1;
	} else {
		res = 0;
	}
	
	if (error != NULL) *error = CDSD_ESUCCESS;
	close(sockfd);
	
	return res;
}

int libcdsd_connection_available(int *error)
{
	int sockfd;
	
	if ((sockfd = settings_client_connection_create()) == -1) {
		if (error != NULL) *error = CDSD_ENOSERVICE;
		return -1;
	}
	
	if (error != NULL) *error = CDSD_ESUCCESS;
	close(sockfd);
	
	return 1;
}

/*int libcdsd_get_error(void)
{
	return cdsderror;
}*/
