/*
 *      settings.c
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */  


#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <grp.h>
#include <pwd.h>
#include <gdbm.h>

#include "settings.h"

static char *settings_get_dirname(char *filename);
static int settings_mkdir(char *dir);

int settings_parse_config_file(const char *filename, const char *section, void *userdata, int(*callback(char *parameter, char *argument, void *userdata)))
{
	FILE *config;
	char string[4096];
	char *del;
	char *parameter;
	char *argument;
	int chr, len;
	int sectionflag = 1;
	
	
	if (callback == NULL) return 0;
	
	if ((config = fopen(filename, "r")) == NULL) return 0;
	
	while (fgets((char *)&string, 4096, config) != NULL) {
		len = strlen(string);
		while ((len > 0) && ((string[len-1] == ' ') || (string[len-1] == '\n'))) len--;
		string[len] = '\0';
		
		chr = 0;
		while ((chr <= len) && ((string[chr] == ' ') || (string[chr] == '\n'))) chr++;
		if ((len <= chr) || (string[chr] == ';') || (string[chr] == '#')) continue;
		
		del = strchr((char *)&string, '=');
		
		if (del == NULL) {
			if (section != NULL) {
				if ((string[chr] == '[') && (strchr(&string[chr+1], ']') != NULL)) {
					string[len-chr-1] = '\0';
					parameter = &string[chr+1];
					if (strcmp(section, parameter) == 0) {
						sectionflag = 1;
					} else {
						sectionflag = 0;
					}
				}
			}
		} else if (del != NULL) {
			len = strlen(string) - strlen(del);
			
			parameter = &string[chr];
			string[len-chr] = '\0';
			
			if ((len = strlen(del+1)) <= 0) continue;
			argument = del+1;
			argument[len] = '\0';
			
			callback(parameter, argument, userdata);
		}
	}
	
	fclose(config);
	
	return 1;
}

int settings_server_connection_create(void)
{
	int sockfd;
	//socklen_t len;
	struct sockaddr_un address;
	//int flags;
	
	if ((sockfd = socket(AF_LOCAL, SOCK_STREAM, 0)) < 0) return -1;
	
	unlink(SETTINGS_SOCKET_FILE);
	
	memset(&address, 0, sizeof(struct sockaddr_un));
	address.sun_family = AF_LOCAL;
	strncpy((char *)&address.sun_path, SETTINGS_SOCKET_FILE, sizeof(address.sun_path)-1);
	bind(sockfd, (const struct sockaddr *)&address, SUN_LEN(&address));
	chmod(address.sun_path, 0777);
	
	/*flags = fcntl(sockfd, F_GETFL, 0);
	fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);*/
	
	if (listen(sockfd, SETTINGS_SOCKET_LISTENQ) < 0) {
		close(sockfd);
		return -1;
	}
	
	return sockfd;
}

void settings_server_connection_close(int sockfd)
{
	close(sockfd);
	unlink(SETTINGS_SOCKET_FILE);
}

int settings_read_credentials(int sockfd, struct _cmd *cmd, char *data, size_t datasize, struct ucred *credentials)
{
	struct msghdr msg;
	struct iovec iovrecv[2];
	struct cmsghdr *msgch;
	struct ucred *cred;
	int msglen;
	const int onflag = 1;
	union {
        struct cmsghdr cmh;
        char   control[CMSG_SPACE(sizeof(struct ucred))];
    } control_un;
		
	
	memset(&msg, 0, sizeof(struct msghdr));
	
	if ((data != NULL) && (datasize > 0) && (cmd != NULL)) {
		//control message with data
		memset(data, 0, datasize);
		memset(cmd, 0, sizeof(struct _cmd));
		iovrecv[0].iov_base = cmd;
		iovrecv[0].iov_len = sizeof(struct _cmd);
		iovrecv[1].iov_base = data;
		iovrecv[1].iov_len = datasize;
		msg.msg_iov = iovrecv;
		msg.msg_iovlen = 2;
	} else if (((data == NULL) || (datasize <= 0)) && (cmd != NULL)) {
		//control message without data
		memset(cmd, 0, sizeof(struct _cmd));
		iovrecv[0].iov_base = cmd;
		iovrecv[0].iov_len = sizeof(struct _cmd);
		msg.msg_iov = iovrecv;
		msg.msg_iovlen = 1;
	} else if ((data != NULL) && (datasize > 0) && (cmd == NULL)) {
		//udev message
		memset(data, 0, datasize);
		iovrecv[0].iov_base = data;
		iovrecv[0].iov_len = datasize;
		msg.msg_iov = iovrecv;
		msg.msg_iovlen = 1;
	} else {
		//empty message
		msg.msg_iov = NULL;
		msg.msg_iovlen = 0;
	}
	
	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	msg.msg_control = control_un.control;
	msg.msg_controllen = sizeof(control_un.control);
	
	control_un.cmh.cmsg_len = CMSG_LEN(sizeof(struct ucred));
	control_un.cmh.cmsg_level = SOL_SOCKET;
	control_un.cmh.cmsg_type = SCM_CREDENTIALS;
	
	setsockopt(sockfd, SOL_SOCKET, SO_PASSCRED, &onflag, sizeof(onflag));
	
	if ((msglen = recvmsg(sockfd, &msg, 0)) == -1) {
		return -1;
	}
	
	//credentials
	msgch = CMSG_FIRSTHDR(&msg);
	if ((msgch != NULL) && (msgch->cmsg_type == SCM_CREDENTIALS) && (credentials != NULL)) {
		cred = (struct ucred *)CMSG_DATA(msgch);
		memcpy(credentials, cred, sizeof(struct ucred));
	}
	
	//return only payload size without command header if one exists
	if (cmd != NULL) msglen = msglen - sizeof(struct _cmd);
	
	return msglen;
}

int settings_client_connection_create(void)
{
	int sockfd;
	struct sockaddr_un address;
	//int flags;
	
	if ((sockfd = socket(AF_LOCAL, SOCK_STREAM, 0)) < 0) return -1;
	
	memset(&address, 0, sizeof(struct sockaddr_un));
	address.sun_family = AF_LOCAL;
	strncpy((char *)&address.sun_path, SETTINGS_SOCKET_FILE, sizeof(address.sun_path)-1);
	
	if (connect(sockfd, (const struct sockaddr *)&address, sizeof(struct sockaddr_un)) == -1) {
		close(sockfd);
		return -1;
	}
	
	/*flags = fcntl(sockfd, F_GETFL, 0);
	fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);*/
	
	return sockfd;
}

int settings_send_credentials(int sockfd, struct _cmd *cmd, char *data, size_t datasize)
{
	struct msghdr msg;
    struct iovec iovsend[2];
    struct ucred *cred;
    struct cmsghdr *msgch;
    int msglen;
    union {
        struct cmsghdr cmh;
        char   control[CMSG_SPACE(sizeof(struct ucred))];
	} control_un;
    
	if ((data != NULL) && (datasize > 0) && (cmd != NULL)) {
		//control message with data
		msg.msg_iov = iovsend;
		msg.msg_iovlen = 2;
		iovsend[0].iov_base = cmd;
		iovsend[0].iov_len = sizeof(struct _cmd);
		iovsend[1].iov_base = data;
		iovsend[1].iov_len = datasize;
	} else if (((data == NULL) || (datasize <= 0)) && (cmd != NULL)) {
		//control message without data
		msg.msg_iov = iovsend;
		msg.msg_iovlen = 1;
		iovsend[0].iov_base = cmd;
		iovsend[0].iov_len = sizeof(struct _cmd);
	} else {
		//empty message
		msg.msg_iov = NULL;
		msg.msg_iovlen = 0;
	}
	
	msg.msg_name = NULL;
	msg.msg_namelen = 0;
	
	msg.msg_control = control_un.control;
	msg.msg_controllen = sizeof(control_un.control);
	
	//credentials
	msgch = CMSG_FIRSTHDR(&msg);
	msgch->cmsg_len = CMSG_LEN(sizeof(struct ucred));
	msgch->cmsg_level = SOL_SOCKET;
	msgch->cmsg_type = SCM_CREDENTIALS;
	cred = (struct ucred *) CMSG_DATA(msgch);
	
	cred->pid = getpid();
	cred->uid = getuid();
	cred->gid = getgid();
	
	if ((msglen = sendmsg(sockfd, &msg, 0)) == -1) {
		return -1;
	}
	
	//return only payload size without command header if one exists
	if (cmd != NULL) msglen = msglen - sizeof(struct _cmd);
	
	return msglen;	
}

gid_t settings_get_group_id(char *groupname)
{
	struct group *grp;
	gid_t gid;
	char *endptr;
	
	if ((groupname == NULL) || (groupname[0] == '\0')) return -1;
	
	errno = 0;
	gid = strtol(groupname, (char **)&endptr, 10);
	
	if (((errno != 0) && (gid == 0)) || (endptr == groupname)) {
		if ((grp = getgrnam(groupname)) != NULL) {
			gid = grp->gr_gid;
		} else {
			gid = -1;
		}
	}
		
	return gid;
}

int settings_user_in_group(int uid, int gid)
{
	int ngroups = 256;
	gid_t *groups;
	struct passwd *pwd;
	int i;
	
	if ((pwd = getpwuid(uid)) == NULL) return 0;
	
	if (pwd->pw_gid == gid) return 1;
	
	if ((groups = calloc(ngroups, sizeof(gid_t))) == NULL) return 0;
	
	if (getgrouplist(pwd->pw_name, pwd->pw_gid, groups, &ngroups) == -1) {
		if ((groups = realloc(groups, ngroups*sizeof(gid_t))) == NULL) {
			free(groups);
			return 0;
		}
		memset(groups, 0, ngroups*sizeof(gid_t));
		if (getgrouplist(pwd->pw_name, pwd->pw_gid, groups, &ngroups) == -1) {
			free(groups);
			return 0;
		}
	}
	
	for (i=0; i<ngroups; i++) {
		if (gid == groups[i]) {
			free(groups);
			return 1;
		}
	}
	free(groups);
	
	return 0;
}

char *settings_string_get_fragment(char *str, char *param, char *term, char *buffer, ssize_t bufsize)
{
	char *first, *second;
	ssize_t linklen;
	
	if ((str == NULL) || (buffer == NULL)) return NULL;
	
	if ((param != NULL) && ((first = strstr(str, param)) != NULL)) {
		first += strlen(param);
	} else return NULL;
		
	if ((second = strstr(first+1, term == NULL ? "\n" : term)) != NULL) {
		if ((linklen = second - first) > 0) {
			if (linklen > bufsize) linklen = bufsize;
			strncpy(buffer, first, linklen);
			buffer[linklen] = '\0';
			
			return buffer;
		}
	}
	
	return NULL;
}

unsigned int settings_string_get_uint(char *str, char *param, char *term)
{
	char *first, *second;
	ssize_t linklen;
	char buffer[32];
	unsigned int value = 0;
	
	if ((str == NULL) || (buffer == NULL)) return value;
	
	if ((param != NULL) && ((first = strstr(str, param)) != NULL)) {
		first += strlen(param);
	} else return value;
	
	if ((second = strstr(first+1, term == NULL ? "\n" : term)) != NULL) {
		if ((linklen = second - first) > 0) {
			if (linklen > sizeof(buffer)) linklen = sizeof(buffer);
			strncpy((char *)&buffer, first, linklen);
			buffer[linklen] = '\0';
			value = strtoul(buffer, (char **)NULL, 10);
			return value;
		}
	}
	
	return value;
}

long int settings_string_get_int(char *str, char *param, char *term)
{
	char *first, *second;
	ssize_t linklen;
	char buffer[32];
	long int value = 0;
	
	if ((str == NULL) || (buffer == NULL)) return value;
	
	if ((param != NULL) && ((first = strstr(str, param)) != NULL)) {
		first += strlen(param);
	} else return value;
	
	if ((second = strstr(first+1, term == NULL ? "\n" : term)) != NULL) {
		if ((linklen = second - first) > 0) {
			if (linklen > sizeof(buffer)) linklen = sizeof(buffer);
			strncpy((char *)&buffer, first, linklen);
			buffer[linklen] = '\0';
			value = strtol(buffer, (char **)NULL, 10);
			return value;
		}
	}
	
	return value;
}

static char *settings_get_dirname(char *filename)
{
	char *fname;
	char *dirname;
	int dirlen;
	
	if (filename == NULL) return NULL;
	
	if ((fname = strrchr(filename, '/')) != NULL) {
		if (fname[1] != '\0') {
			if ((dirlen = strlen(filename) - strlen(fname)) > 0) {
				if ((dirname = calloc(1, dirlen)) != NULL) {
					strncpy(dirname, filename, dirlen);
					dirname[dirlen] = '\0';
					
					return dirname;
				}
			}
		}
	}
	
	return NULL;
}

static int settings_mkdir(char *dir)
{
	struct stat statbuf;
	
	if (dir == NULL) return 0;
	
	if (stat(dir, &statbuf) == -1) {
		if (errno == ENOENT) {
			if (mkdir((const char *)dir, S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IROTH) != -1) {
				return 1;
			}
		}
	} else if (S_ISDIR(statbuf.st_mode)) {
		return 1;
	}
		
	return 0;
}

GDBM_FILE settings_db_open(char *filename)
{
	GDBM_FILE dbf;
	char *dirname;
	
	if ((dirname = settings_get_dirname(filename)) == NULL) return NULL;	
	
	if (settings_mkdir(dirname)) {
		if ((dbf = gdbm_open(filename, 0, GDBM_WRCREAT, 0755, 0)) != NULL) {
			free(dirname);
			return dbf;
		}
	}
	free(dirname);
	
	return NULL;
}

void settings_db_close(GDBM_FILE dbf)
{
	if (dbf != NULL) {
		gdbm_sync(dbf);
		gdbm_close(dbf);
	}
}

int settings_db_fetch(GDBM_FILE dbf, struct optical_device *device)
{
	char devid[45];
	int len;
	datum key, data;
	
	if ((dbf == NULL) || (device == NULL)) return 0;
	
	len = snprintf((char *)&devid, sizeof(devid), "%s_%s_%s", device->vendorid, device->productid, device->productver);
	devid[len] = '\0';
	
	if (len == 0) return 0;
	
	key.dptr = strdup(devid);
	key.dsize = len+1;
	
	if (!gdbm_exists(dbf, key)) {
		free(key.dptr);
		return 0;
	}
	
	data = gdbm_fetch(dbf, key);
	
	if (data.dptr != NULL) {
		device->user_cd_speed = (unsigned short)settings_string_get_uint(data.dptr, "cd=\"", "\"");
		device->user_dvd_speed = (unsigned short)settings_string_get_uint(data.dptr, "dvd=\"", "\"");
		device->user_bd_speed = (unsigned short)settings_string_get_uint(data.dptr, "bd=\"", "\"");
		device->user_polling = (unsigned short)settings_string_get_uint(data.dptr, "poll=\"", "\"");
		free(key.dptr);
		free(data.dptr);
		return 1;
	}
	
	free(key.dptr);
	
	return 0;
}

int settings_db_store(GDBM_FILE dbf, struct optical_device *device)
{
	char devid[45];
	char devxml[1024];
	int len;
	datum key, data;
	
	if ((dbf == NULL) || (device == NULL)) return 0;
	
	len = snprintf((char *)&devid, sizeof(devid), "%s_%s_%s", device->vendorid, device->productid, device->productver);
	devid[len] = '\0';
	
	if (len == 0) return 0;
	
	key.dptr = strdup(devid);
	key.dsize = len+1;
	
	len = snprintf((char *)&devxml, sizeof(devxml), SETTINGS_DB_ENTRY, device->user_cd_speed, device->user_dvd_speed, device->user_bd_speed, device->user_polling, devid);
	devxml[len] = '\0';
	
	data.dptr = strdup(devxml);
	data.dsize = len+1;
	
	if (gdbm_store(dbf, key, data, GDBM_REPLACE) != 0) {
		free(key.dptr);
		free(data.dptr);
		return 0;
	}
	
	gdbm_sync(dbf);
	free(key.dptr);
	free(data.dptr);
	
	return 1;
}
