/*
 *      udevclient.c
 *      
 *      Copyright 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/socket.h>
#include <string.h>
#include <sys/poll.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <errno.h>
#include <dirent.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <ctype.h>
#include <linux/types.h>
#include <linux/netlink.h>
#include <linux/netlink.h>

/* References:
 *
 * libudev - interface to udev device information
 *
 * Copyright (C) 2008-2010 Kay Sievers <kay.sievers@vrfy.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 */

#define UDEV_MONITOR_MAGIC          0xfeedcafe

enum {
	UDEVCLIENT_SOURCE_KERNEL = 0,
	UDEVCLIENT_SOURCE_LIB
};

struct udev_monitor_netlink_header {
      /* "libudev" prefix to distinguish libudev and kernel messages */
      char prefix[8];
      /*
       * magic to protect against daemon <-> library message format mismatch
       * used in the kernel from socket filter rules; needs to be stored in network order
       */
      unsigned int magic;
      /* total length of header structure known to the sender */
      unsigned int header_size;
      /* properties string buffer */
      unsigned int properties_off;
      unsigned int properties_len;
      /*
       * hashes of primary device properties strings, to let libudev subscribers
       * use in-kernel socket filters; values need to be stored in network order
       */
      unsigned int filter_subsystem_hash;
      unsigned int filter_devtype_hash;
      unsigned int filter_tag_bloom_hi;
      unsigned int filter_tag_bloom_lo;
};

static int udevclient_test_descriptor(int fd);

static int udevclient_test_descriptor(int fd)
{
	struct stat statbuf;
	
	if (fstat(fd, &statbuf) < 0) return 0;
	
	if (statbuf.st_mode & S_IFSOCK) return 1; else return 0;
}

int udevclient_connection_create(void)
{
	int socketfd;
	struct sockaddr_nl address;
	const int enabled = 1;
	
	memset(&address, 0, sizeof(struct sockaddr_nl));
	address.nl_family = AF_NETLINK;
	address.nl_pid = getpid();
	address.nl_groups = -1;
	
	if ((socketfd = socket(PF_NETLINK, SOCK_DGRAM | SOCK_CLOEXEC, NETLINK_KOBJECT_UEVENT)) == -1) {
		perror("[-] Failed to create netlink socket");
		return -1;
	}
		
	if (bind(socketfd, (void *)&address, sizeof(struct sockaddr_nl)) == -1) {
		perror("[-] Failed to bind netlink socket");
		close(socketfd);
		return -1;
	}
	
	if (setsockopt(socketfd, SOL_SOCKET, SO_PASSCRED, &enabled, sizeof(enabled)) == -1) {
		perror("[-] Failed to set netlink socket options");
		close(socketfd);
		return -1;
	}
	
	return socketfd;	
}

void udevclient_connection_close(int fd)
{
	if (udevclient_test_descriptor(fd)) close(fd);
}

char *udevclient_get_event_string(char *str, char *param, char *buffer, ssize_t bufsize)
{
	char *first, *second;
	ssize_t linklen;
	
	if ((str == NULL) || (buffer == NULL) || (param == NULL)) return NULL;
	
	if ((param != NULL) && ((first = strstr(str, param)) != NULL)) {
		first += strlen(param);
	}
	
	if ((second = strstr(first+1, "\n")) != NULL) {
		if ((linklen = second - first) > 0) {
			if (linklen > bufsize) linklen = bufsize;
			strncpy(buffer, first, linklen);
			buffer[linklen] = '\0';
			
			return buffer;
		}
	}
	
	return NULL;
}

int udevclient_validate_change_block_device_event(char *request, unsigned int *major, unsigned int *minor)
{
	char parambuf[1024];
	char *param;
	
		
	if (((param = udevclient_get_event_string(request, "ACTION=", (char *)&parambuf, sizeof(parambuf))) == NULL) || (strcmp(param, "change") != 0)) return 0;
	if (((param = udevclient_get_event_string(request, "SUBSYSTEM=", (char *)&parambuf, sizeof(parambuf))) == NULL) || (strcmp(param, "block") != 0)) return 0;
	
	if ((param = udevclient_get_event_string(request, "MAJOR=", (char *)&parambuf, sizeof(parambuf))) != NULL) {
		*major = atoi(param);
	} else {
		return 0;
	}
	
	if ((param = udevclient_get_event_string(request, "MINOR=", (char *)&parambuf, sizeof(parambuf))) != NULL) {
		*minor = atoi(param);
	} else {
		return 0;
	}
	
	return 1;
}

void udevclient_parse_request(char *message, int msglen, int (*event_cb)(unsigned int major, unsigned int minor))
{
	int i, r, rlen;
	struct udev_monitor_netlink_header udevhdr;
	int source;
	char *request;
	unsigned int major, minor;
	
	if (strncmp((const char *)message, "libudev", 7) == 0) {
		memcpy(&udevhdr, message, sizeof(struct udev_monitor_netlink_header));
		//if (udevhdr.magic != htonl(UDEV_MONITOR_MAGIC)) return;
		source = UDEVCLIENT_SOURCE_LIB;
		i = sizeof(struct udev_monitor_netlink_header);
		request = (char *)calloc(1, msglen - i);
	} else {
		i = 0;
		source = UDEVCLIENT_SOURCE_KERNEL;
		request = (char *)calloc(1, msglen);
	}
		
	if (source == UDEVCLIENT_SOURCE_KERNEL) {
		r = 0;
		while (i < msglen) {
			if ((rlen = strlen(message + i)) > 0) {
				strncpy(request + r, message + i, rlen);
				request[r+rlen] = '\n';
				i += rlen + 1;
				r += rlen + 1;
			}
		}
		
		major = minor = 0;
		if (udevclient_validate_change_block_device_event(request, &major, &minor)) {
			if (event_cb != NULL) event_cb(major, minor);
		}
		
		free(request);
	}
}
